/*
 * Copyright (c) 2004 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: sm-conf-prtcnf.c,v 1.10 2005/06/20 19:03:15 ca Exp $")

#include "sm/string.h"
#include "sm/sm-conf.h"
#include "sm/sm-conf-prt.h"
#include "sm-conf-type.h"
#include "sm/io.h"
#include "sm/str.h"
#include "sm/util.h"
#include "prtcnf.h"

/*
**  Show configuration data.
**  This shows the data that is actually read from a configuration file
**  and stored in a structure.
*/

/*
**  SM_CONF_SUB2DEF -- convert a value into its string representation
**
**	Parameters:
**		def -- sentinel-terminated array of subdefinitions.
**		type -- if non-NULL, only entries of this type are
**			taken into account
**		val -- (choice) value to convert
**		fp -- file for output
**
**	Returns:
**		SM_SUCCESS
*/

static int
sm_conf_sub2valbit(sm_conf_definition_T const *def, sm_conf_type_T const *type, uint32_t val, sm_file_T *fp)
{
	SM_IS_CONF_DEF(def);
	while (def->scd_name != NULL)
	{
		if ((type == NULL || def->scd_type == type) &&
		    (val & def->scd_offset) != 0)
			sm_io_fprintf(fp, "%s, ", def->scd_name);
		def++;
	}
	return SM_SUCCESS;
}

/*
**  SM_CONF_SUB2DEF -- convert a value into its string representation
**
**	Parameters:
**		def -- sentinel-terminated array of subdefinitions.
**		type -- if non-NULL, only entries of this type are
**			taken into account
**		val -- (choice) value to convert
**		fp -- file for output
**
**	Returns:
**		SM_SUCCESS
*/

static int
sm_conf_sub2valeq(sm_conf_definition_T const *def, sm_conf_type_T const *type, uint32_t val, sm_file_T *fp)
{
	SM_IS_CONF_DEF(def);
	while (def->scd_name != NULL)
	{
		if ((type == NULL || def->scd_type == type) &&
		    val == def->scd_offset)
		{	
			sm_io_fprintf(fp, "%s", def->scd_name);
			break;
		}
		def++;
	}
	return SM_SUCCESS;
}

/*
**  SM_CONF_PRT_CNF -- print content of one node
**
**	Parameters:
**		def -- configuration definition.
**		cnf -- pointer to structure holding the configuration data
**		offset -- current offset in struct (cnf)
**		fp -- file for output
**		indent -- current indentation
**		prefix -- prefix to print (if not NULL)
**
**	Returns:
**		SM_SUCCESS
*/

#define SM_C_DELIM	'.'
#define SM_S_OPEN	'['
#define SM_S_CLOSE	']'

static int
sm_conf_prt_cnf(const sm_conf_definition_T *def, const void *cnf, ptrdiff_t offset, sm_file_T *fp, int indent, sm_str_P prefix)
{
	const sm_conf_type_T *scdtype;
	ptrdiff_t off;
	char *s;

	SM_IS_CONF_DEF(def);
	scdtype = def->scd_type;
	if (scdtype != NULL)
	{
#define PRT_COMMENT	\
	do		\
	{		\
		if (def->scd_description != NULL &&	\
		    *(def->scd_description) != '\0')	\
		{					\
			prt_fmtted(fp, indent, "# ", def->scd_description); \
		}					\
	} while (0)

		off = offset + def->scd_offset;

#define CNF_VAL	(((char *)cnf) + off)

		if (scdtype == sm_conf_type_section)
		{
			PRT_COMMENT;
			if (prefix != NULL)
			{
				sm_ret_T ret;
				size_t oldlen;

				oldlen = sm_str_getlen(prefix);
				ret = sm_str_scat(prefix, def->scd_name);
				if (sm_is_err(ret))
					return ret;

#if 0
				if (def->scn_section.scns_keyword != NULL)
				{
					ret = sm_str_scat0(prefix,
							def->scd_name);
					if (sm_is_err(ret))
						return ret;
				}
#endif /* 0 */

				ret = sm_str_put(prefix, SM_C_DELIM);
				if (sm_is_err(ret))
					return ret;
				sm_conf_prt_cnfs(def->scd_contents, cnf,
					off, fp, indent + 2, prefix);
				SM_STR_SETLEN(prefix, oldlen);
			}
			else
			{
				sm_io_fprintf(fp, "\n%*s%s {\n"
					, indent, ""
					, def->scd_name);
				sm_conf_prt_cnfs(def->scd_contents, cnf, off,
					fp, indent + 2, prefix);
				sm_io_fprintf(fp, "%*s}\n"
					, indent, "");
			}
		}
		else if (scdtype == sm_conf_type_string)
		{
			s = (char *)*((char **)CNF_VAL);
			if (s != NULL)
			{
				PRT_COMMENT;
				sm_io_fprintf(fp, "%*s", indent, "");
				if (prefix != NULL)
					sm_io_fprintf(fp, "%S", prefix);
				sm_io_fprintf(fp, "%s=\"%s\";\n"
						, def->scd_name
						, s);
			}
		}
		else if (scdtype == sm_conf_type_char)
		{
			uchar c;

			c = *((uchar *)CNF_VAL);
			if (c != '\0')
			{
				PRT_COMMENT;
				sm_io_fprintf(fp, "%*s", indent, "");
				if (prefix != NULL)
					sm_io_fprintf(fp, "%S", prefix);
				sm_io_fprintf(fp, "%s=\"%c\";\n"
					, def->scd_name
					, c);
			}
		}
		else if (scdtype == sm_conf_type_u32)
		{
			PRT_COMMENT;
			sm_io_fprintf(fp, "%*s", indent, "");
			if (prefix != NULL)
				sm_io_fprintf(fp, "%S", prefix);
			if (def->scd_contents != NULL &&
			    def->scd_contents->scd_type ==
				sm_conf_type_u32_suffix &&
			    def->scd_size == sizeof(uint32_t))
			{
				sm_io_fprintf(fp, "%s=", def->scd_name);
				sm_u32totxtwsuffix(def, *(uint32_t *)CNF_VAL,
					fp);
				sm_io_fprintf(fp, ";\n");
			}
			else if (def->scd_size == sizeof(uint32_t))
				sm_io_fprintf(fp, "%s=%lu;\n"
					, def->scd_name
					, (unsigned long)*(uint32_t *)CNF_VAL);
			else if (def->scd_size == sizeof(uint16_t))
				sm_io_fprintf(fp, "%s=%u;\n"
					, def->scd_name
					, (unsigned int)*(uint16_t *)CNF_VAL);
			else if (def->scd_size == sizeof(uint8_t))
				sm_io_fprintf(fp, "%s=%hu;\n"
					, def->scd_name
					, (unsigned short)*(uint8_t *)CNF_VAL);
		}
		else if (scdtype == sm_conf_type_ipv4)
		{
			PRT_COMMENT;
			sm_io_fprintf(fp, "%*s", indent, "");
			if (prefix != NULL)
				sm_io_fprintf(fp, "%S", prefix);
			sm_io_fprintf(fp, "%s=%A;\n"
				, def->scd_name
				, *(uint32_t *)CNF_VAL);
		}
		else if (scdtype == sm_conf_type_argv)
		{
			size_t i;
			char **argv;

			argv = (char **)CNF_VAL;
			if (argv != NULL && argv[0] != NULL)
			{
				PRT_COMMENT;
				sm_io_fprintf(fp, "%*s%s={"
					, indent, ""
					, def->scd_name);
				for (i = 0; (s = argv[i]) != NULL; i++)
				{
					sm_io_fprintf(fp, "%s, ", s);
				}
				sm_io_fprintf(fp, "}\n");
			}
		}
		else if (scdtype == sm_conf_type_choice)
		{
			uint32_t u;

			u = *(uint32_t *)CNF_VAL;
			if (u != 0 &&
			    (def->scd_flags & SM_CONF_FLAG_MULTIPLE) != 0)
			{
				PRT_COMMENT;
				sm_io_fprintf(fp, "%*s%s={"
					, indent, ""
					, def->scd_name);
				(void) sm_conf_sub2valbit(def->scd_contents,
						NULL, u, fp);
				sm_io_fprintf(fp, "}\n");
			}
			else if (u != 0)
			{
				PRT_COMMENT;
				sm_io_fprintf(fp, "%*s%s="
					, indent, ""
					, def->scd_name);
				(void) sm_conf_sub2valeq(def->scd_contents,
						NULL, u, fp);
				sm_io_fprintf(fp, ";\n");
			}
		}
		else if (scdtype == sm_conf_type_array)
		{
			const sm_conf_definition_T *array = def->scd_contents;

			if (array[0].scd_name != NULL &&
			    array[0].scd_type == sm_conf_type_section &&
			    array[1].scd_name != NULL &&
			    array[1].scd_type == sm_conf_type_array_n
			   )
			{
				ptrdiff_t offn;
				unsigned int n, i;

				PRT_COMMENT;
				offn = offset + array[1].scd_offset;
				n = *(unsigned int *) (((char *)cnf) + offn);
				for (i = 0; i < n; i++)
				{
					sm_io_fprintf(fp, "\n%*s%s {\n"
						, indent, ""
						, def->scd_name);
					sm_conf_prt_cnfs(array[0].scd_contents,
						cnf,
						offset + array[0].scd_offset
						+ i * array[0].scd_size,
						fp, indent + 2, prefix);
					sm_io_fprintf(fp, "%*s}\n"
						, indent, "");
				}
			}
#if 0
			else if (array[0].scd_name != NULL &&
			    array[0].scd_type == sm_conf_type_union &&
			    array[1].scd_name != NULL &&
			    array[1].scd_type == sm_conf_type_array_n
			   )
			{
				ptrdiff_t offn;
				unsigned int n, i;

				PRT_COMMENT;
	/* XXX How to print a union? */
				offn = offset + array[1].scd_offset;
				n = *(unsigned int *) (((char *)cnf) + offn);
				for (i = 0; i < n; i++)
				{
					sm_io_fprintf(fp, "\n%*s%s {\n"
						, indent, ""
						, def->scd_name);
					sm_conf_prt_cnfs(array[0].scd_contents,
						cnf,
						offset + array[0].scd_offset
						+ i * array[0].scd_size,
						fp, indent + 2);
					sm_io_fprintf(fp, "%*s}\n"
						, indent, "");
				}
			}
#endif /* 0 */
		}
	}
	return 0;
}

/*
**  SM_CONF_PRT_CNFS -- print content of configuration tree
**
**	Parameters:
**		def -- configuration definition.
**		cnf -- pointer to structure holding the configuration data
**		offset -- current offset in struct (cnf)
**		fp -- file for output
**		indent -- current indentation
**		prefix -- prefix to print (if not NULL)
**
**	Returns:
**		SM_SUCCESS
*/

int
sm_conf_prt_cnfs(const sm_conf_definition_T *defs, const void *cnf, ptrdiff_t offset, sm_file_T *fp, int indent, sm_str_P prefix)
{
	const sm_conf_definition_T *def;

	while ((def = defs++) != NULL && def->scd_name != NULL)
	{
		SM_IS_CONF_DEF(def);
		sm_conf_prt_cnf(def, cnf, offset, fp, indent, NULL);
	}
	return 0;
}

/*
**  SM_CONF_PRT_CONF -- print content of configuration tree
**	(wrapper for sm_conf_prt_cnfs())
**
**	Parameters:
**		def -- configuration definition.
**		cnf -- pointer to structure holding the configuration data
**		fp -- file for output
**
**	Returns:
**		SM_SUCCESS
*/

int
sm_conf_prt_conf(const sm_conf_definition_T *defs, const void *cnf, sm_file_T *fp)
{
	return sm_conf_prt_cnfs(defs, cnf, 0, fp, 0, NULL);
}
