/*
 * Copyright (c) 2002, 2005 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: unixsock.c,v 1.11 2005/03/24 19:55:21 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/fcntl.h"
#include "sm/unixsock.h"

/*
**  UNIX_SERVER_LISTEN_ADDR -- Create a socket, bind, and listen.
**
**	Parameters:
**		my_addr -- sockaddr listen on
**		addrlen -- Length of my_addr
**		backlog -- Backlog of connections to accept
**
**	Returns:
**		New socket fd or sm_error
*/

sm_ret_T
unix_server_listen_addr(sockaddr_un_T *my_addr, socklen_T addrlen, int backlog)
{
	int listenfd;
#if 0
	int sockopt;
#endif
	sm_ret_T ret;

	SM_REQUIRE(my_addr != NULL);
	SM_REQUIRE(backlog > 0);

	/* Open listen port */
	listenfd = socket(AF_UNIX, SOCK_STREAM, 0);
	if (listenfd == -1)
		return sm_error_perm(SM_EM_NET, errno);
#if 0
	sockopt = 1;
	if (setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, (void *) &sockopt,
		       sizeof(sockopt)) == -1)
	{
		ret = sm_error_perm(SM_EM_NET, errno);
		goto error;
	}
#endif /* 0 */

	if (bind(listenfd, (sockaddr_T *) my_addr, addrlen) == -1)
	{
		ret = sm_error_perm(SM_EM_NET, errno);
		goto error;
	}

	if (listen(listenfd, backlog) == -1)
	{
		ret = sm_error_perm(SM_EM_NET, errno);
		goto error;
	}
	return listenfd;

 error:
	close(listenfd);
	return ret;
}

/*
**  UNIX_SERVER_ACCEPT -- Accept an incoming network connection.
**
**	Parameters:
**		listenfd -- fd that we are listening on.
**		addr -- client address that is connecting.
**		addrlen -- length of addr
**	Returns:
**		New socket fd or sm_error
*/

sm_ret_T
unix_server_accept(int listenfd, struct sockaddr *addr, sockaddr_len_T *addrlen)
{
	int connfd;

	SM_REQUIRE(listenfd >= 0);
	SM_REQUIRE(addr != NULL);
	SM_REQUIRE(addrlen != NULL);

	if ((connfd = accept(listenfd, addr, addrlen)) == -1)
	{
		if (errno != EAGAIN)
			return sm_error_perm(SM_EM_NET, errno);
	}
#if 0
	else
	{
		int sockopt;

		sockopt = 1;
		if (setsockopt(connfd, SOL_SOCKET, SO_KEEPALIVE,
			       (void *) &sockopt, sizeof(sockopt)) == -1)
			return sm_error_perm(SM_EM_NET, errno);
	}
#endif /* 0 */
	return connfd;
}

#if 0

/*
**  UNIX_SOCKET_CREATE -- Create a buffered NetSocket object from an fd.
**
**	Creates a buffered NetSocket object suitable for
**	use with below calls.  Sets the O_NONBLOCK fcntl option.
**
**	Parameters
**		fd -- fd to create object from.
**
**	Returns:
**		A NetSocket object
*/
int
unix_socket_create(int fd)
{
	int sock;
	int flags;

	SM_REQUIRE(fd >= 0);
	sock = fd;
	if ((flags = fcntl(fd, F_GETFL, 0)) == -1)
		return sm_error_perm(SM_EM_NET, errno);
	if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1)
		return sm_error_perm(SM_EM_NET, errno);
	return sock;
}
#endif /* 0 */

#if 0

/*
**  UNIX_OUT -- Write len bytes from src to fd.
**
**	Write len bytes from src to fd.  Blocks until all data is transmitted.
**	Will always transmit full len of data (this differs from send(2), which
**	may return less than len).
**
**	Parameters:
**		outfd -- fd to write to
**		src -- data to send
**		len -- length of data to send
**
**	Returns:
**		Number of characters sent or -1 on error.
*/

/*
**  NT does send so that it blocks until all data is away, Unix doesn't, so
**  we have to check the return on Unix.
*/

ssize_t
unix_out(int outfd, const uchar *src, size_t len)
{
	ssize_t l, ret;

	SM_REQUIRE(outfd >= 0);
	SM_REQUIRE(src != NULL);

	l = 0;
#ifdef XP_UNIX
	/*
	**  Unix optimization:
	**  Since Unix properly notices sends to closed sockets (vs NT
	**  which blocks forever on a closed socket).
	**  For Unix send it, notice if it's block, and make sure it all
	**  went (this cuts the code down by 75% on 75% of the calls).
	*/

	ret = send(outfd, (char *) src, len, 0);
	if (ret == -1 && errno == EPIPE)
	{
		/* Connection is closed */
		return -1;
	}

	if (ret >= 0)
		l += ret;
#endif /* XP_UNIX */

	while (l < len)
	{
/* XXX use write_wait() */
		/* Use select to wait for the socket to open up */
		fd_set writefds;

		SM_ASSERT(outfd < FD_SETSIZE);
		FD_ZERO(&writefds);
		FD_SET(outfd, &writefds);

		/* XXX - We should probably use poll here */
		if (select(outfd + 1, NULL, &writefds, NULL, NULL) != 1)
			return sm_error_perm(SM_EM_NET, errno);

		/* It should now be clear to write some more */
		ret = send(outfd, (char *) (src + l), len - l, 0);
		if (ret == -1)
			return sm_error_perm(SM_EM_NET, errno);

		if (ret > 0)
			l += ret;
	}

	SM_ASSERT(l != -1);
	return l;
}
#endif /* 0 */
